@@ -3,7 +3,7 @@
 import gi
 gi.require_version('CDesktopEnums', '3.0')
 
-from gi.repository import Gtk, Gdk, CScreensaver, CDesktopEnums, GObject
+from gi.repository import GLib, Gtk, Gdk, CScreensaver, CDesktopEnums, GObject
 import random
 
 import status
@@ -19,6 +19,7 @@ from osk import OnScreenKeyboard
 from floating import ALIGNMENTS
 from util import utils, trackers, settings
 from util.eventHandler import EventHandler
+from util.utils import DEBUG
 
 class Stage(Gtk.Window):
     """
@@ -34,6 +35,10 @@ class Stage(Gtk.Window):
     The Stage reponds pretty much only to the instructions of the
     ScreensaverManager.
     """
+    __gsignals__ = {
+        'needs-refresh': (GObject.SignalFlags.RUN_LAST, None, ()),
+    }
+
     def __init__(self, manager, away_message):
         if status.InteractiveDebug:
             Gtk.Window.__init__(self,
@@ -55,9 +60,10 @@ class Stage(Gtk.Window):
         self.destroying = False
 
         self.manager = manager
-        status.screen = CScreensaver.Screen.new(status.Debug)
         self.away_message = away_message
 
+        self.activate_callback = None
+
         self.monitors = []
         self.last_focus_monitor = -1
         self.overlay = None
@@ -66,8 +72,7 @@ class Stage(Gtk.Window):
         self.unlock_dialog = None
         self.audio_panel = None
         self.info_panel = None
-
-        self.stage_refresh_id = 0
+        self.osk = None
 
         self.floaters = []
 
@@ -88,7 +93,7 @@ class Stage(Gtk.Window):
                         Gdk.EventMask.FOCUS_CHANGE_MASK)
 
         c = Gdk.RGBA(0, 0, 0, 0)
-        self.override_background_color (Gtk.StateFlags.NORMAL, c);
+        self.override_background_color(Gtk.StateFlags.NORMAL, c)
 
         self.update_geometry()
 
@@ -153,13 +158,9 @@ class Stage(Gtk.Window):
         Wallpapers are secondary.
         """
 
-        if status.Debug:
-            print("Stage: Received screen size-changed signal, refreshing stage")
-
-        self.update_geometry()
-        self.move_onscreen()
-        self.overlay.queue_resize()
+        DEBUG("Stage: Received screen size-changed signal, refreshing stage")
 
+        self.emit("needs-refresh")
 
     def on_monitors_changed(self, screen, data=None):
         """
@@ -167,65 +168,21 @@ class Stage(Gtk.Window):
         as on_screen_size_changed), and follow up at idle with actual monitor view
         refreshes (wallpapers.)
         """
-        if status.Debug:
-            print("Stage: Received screen monitors-changed signal, refreshing stage")
+        DEBUG("Stage: Received screen monitors-changed signal, refreshing stage")
 
-        self.update_geometry()
-        self.move_onscreen()
-        self.overlay.queue_resize()
-
-        Gdk.flush()
-
-        self.queue_refresh_stage()
+        self.emit("needs-refresh")
 
     def on_composited_changed(self, screen, data=None):
         if self.get_realized():
+            DEBUG("Stage: Received screen composited-changed signal, refreshing stage")
 
-            user_time = self.get_display().get_user_time()
-
-            self.hide()
-            self.unrealize()
-
-            self.realize()
-
-            self.get_window().set_user_time(user_time)
-            self.show()
-
-            GObject.idle_add(self.manager.grab_stage)
+            self.emit("needs-refresh")
 
     def on_grab_broken_event(self, widget, event, data=None):
         GObject.idle_add(self.manager.grab_stage)
 
         return False
 
-    def queue_refresh_stage(self):
-        """
-        Queues a complete refresh of the stage, resizing the screen if necessary,
-        reconstructing the individual monitor objects, etc...
-        """
-        if self.stage_refresh_id > 0:
-            GObject.source_remove(self.stage_refresh_id)
-            self.stage_refresh_id = 0
-
-        self.stage_refresh_id = GObject.idle_add(self._update_full_stage_on_idle)
-
-    def _update_full_stage_on_idle(self, data=None):
-        self.stage_refresh_id = 0
-
-        self._refresh()
-
-        return False
-
-    def _refresh(self):
-        Gdk.flush()
-        if status.Debug:
-            print("Stage: refresh callback")
-
-        self.update_geometry()
-        self.move_onscreen()
-        self.update_monitors()
-        self.overlay.queue_resize()
-
     def activate(self, callback):
         """
         This is the primary way of making the Stage visible.
@@ -235,7 +192,10 @@ class Stage(Gtk.Window):
         self.move_onscreen()
         self.show()
 
-        callback()
+        if self.get_realized():
+            callback()
+        else:
+            self.activate_callback = callback
 
     def deactivate(self, callback):
         """
@@ -255,14 +215,19 @@ class Stage(Gtk.Window):
         window = self.get_window()
         utils.override_user_time(window)
 
+        CScreensaver.Screen.set_net_wm_name(window, "cinnamon-screensaver-window")
+
         self.setup_children()
 
-        self.gdk_filter.start()
+        self.gdk_filter.start(singletons.MuffinClient.get_using_fractional_scaling(), status.Debug)
 
         trackers.con_tracker_get().disconnect(self.overlay,
                                               "realize",
                                               self.on_realized)
 
+        if self.activate_callback is not None:
+            self.activate_callback()
+
     def move_onscreen(self):
         w = self.get_window()
 
@@ -307,29 +272,32 @@ class Stage(Gtk.Window):
                 print("Problem setting up clock widget: %s" % str(e))
                 self.clock_widget = None
 
-            try:
-                self.setup_albumart()
-            except Exception as e:
-                print("Problem setting up albumart widget: %s" % str(e))
-                self.albumart_widget = None
-
-            try:
-                self.setup_status_bars()
-            except Exception as e:
-                print("Problem setting up status bars: %s" % str(e))
-                self.audio_panel = None
-                self.info_panel = None
-
             try:
                 self.setup_osk()
             except Exception as e:
                 print("Problem setting up on-screen keyboard: %s" % str(e))
                 self.osk = None
 
-        if total_failure:
+            trackers.timer_tracker_get().start("setup-delayed-components",
+                                               2000,
+                                               self.setup_delayed_components)
+        else:
             print("Total failure somewhere, deactivating screensaver.")
             GObject.idle_add(self.deactivate_after_timeout)
 
+    def setup_delayed_components(self, data=None):
+        try:
+            self.setup_albumart()
+        except Exception as e:
+            print("Problem setting up albumart widget: %s" % str(e))
+            self.albumart_widget = None
+        try:
+            self.setup_status_bars()
+        except Exception as e:
+            print("Problem setting up status bars: %s" % str(e))
+            self.audio_panel = None
+            self.info_panel = None
+
     def destroy_children(self):
         try:
             self.destroy_monitor_views()
@@ -337,39 +305,39 @@ class Stage(Gtk.Window):
             print(e)
 
         try:
-            if self.unlock_dialog != None:
+            if self.unlock_dialog is not None:
                 self.unlock_dialog.destroy()
         except Exception as e:
             print(e)
 
         try:
-            if self.clock_widget != None:
+            if self.clock_widget is not None:
                 self.clock_widget.stop_positioning()
                 self.clock_widget.destroy()
         except Exception as e:
             print(e)
 
         try:
-            if self.albumart_widget != None:
+            if self.albumart_widget is not None:
                 self.albumart_widget.stop_positioning()
                 self.albumart_widget.destroy()
         except Exception as e:
             print(e)
 
         try:
-            if self.info_panel != None:
+            if self.info_panel is not None:
                 self.info_panel.destroy()
         except Exception as e:
             print(e)
 
         try:
-            if self.info_panel != None:
+            if self.info_panel is not None:
                 self.audio_panel.destroy()
         except Exception as e:
             print(e)
 
         try:
-            if self.osk != None:
+            if self.osk is not None:
                 self.osk.destroy()
         except Exception as e:
             print(e)
@@ -390,6 +358,27 @@ class Stage(Gtk.Window):
         Performs all tear-down necessary to destroy the Stage, destroying
         all children in the process, and finally destroying itself.
         """
+
+        trackers.con_tracker_get().disconnect(self.unlock_dialog,
+                                              "inhibit-timeout",
+                                              self.set_timeout_active)
+
+        trackers.con_tracker_get().disconnect(self.unlock_dialog,
+                                              "uninhibit-timeout",
+                                              self.set_timeout_active)
+
+        trackers.con_tracker_get().disconnect(self.unlock_dialog,
+                                              "authenticate-success",
+                                              self.authentication_result_callback)
+
+        trackers.con_tracker_get().disconnect(self.unlock_dialog,
+                                              "authenticate-failure",
+                                              self.authentication_result_callback)
+
+        trackers.con_tracker_get().disconnect(self.unlock_dialog,
+                                              "authenticate-cancel",
+                                              self.authentication_cancel_callback)
+
         trackers.con_tracker_get().disconnect(singletons.Backgrounds,
                                               "changed",
                                               self.on_bg_changed)
@@ -404,6 +393,7 @@ class Stage(Gtk.Window):
 
         self.set_timeout_active(None, False)
 
+        trackers.timer_tracker_get().cancel("setup-delayed-components")
         self.destroy_children()
 
         self.gdk_filter.stop()
@@ -422,7 +412,6 @@ class Stage(Gtk.Window):
                                               self.position_overlay_child)
 
         self.destroy()
-        status.screen = None
 
     def setup_monitors(self):
         """
@@ -476,10 +465,10 @@ class Stage(Gtk.Window):
         Callback for UPower changes, this will make our MonitorViews update
         themselves according to user setting and power state.
         """
-        if status.Debug:
-            print("stage: Power state changed, updating info panel")
+        DEBUG("stage: Power state changed, updating info panel")
 
-        self.info_panel.update_visibility()
+        if self.info_panel is not None:
+            self.info_panel.update_visibility()
 
     def setup_clock(self):
         """
@@ -510,7 +499,7 @@ class Stage(Gtk.Window):
         self.albumart_widget = AlbumArt(None, status.screen.get_mouse_monitor())
         self.add_child_widget(self.albumart_widget)
 
-        self.floaters.append(self.clock_widget)
+        self.floaters.append(self.albumart_widget)
 
         if settings.get_show_albumart():
             self.albumart_widget.start_positioning()
@@ -565,10 +554,13 @@ class Stage(Gtk.Window):
         self.audio_panel = AudioPanel()
         self.add_child_widget(self.audio_panel)
 
+        if status.Awake:
+            self.audio_panel.show_panel()
+
         self.info_panel = InfoPanel()
         self.add_child_widget(self.info_panel)
 
-        self.info_panel.update_visibility()
+        self.info_panel.refresh_power_state()
 
     def queue_dialog_key_event(self, event):
         """
@@ -606,7 +598,7 @@ class Stage(Gtk.Window):
         Go back to Sleep if we hit our timer limit
         """
         self.set_timeout_active(None, False)
-        self.manager.cancel_unlock_widget()
+        self.manager.cancel_unlocking()
 
         return False
 
@@ -617,9 +609,9 @@ class Stage(Gtk.Window):
         widget, depending on the outcome.
         """
         if success:
-            if self.clock_widget != None:
+            if self.clock_widget is not None:
                 self.clock_widget.hide()
-            if self.albumart_widget != None:
+            if self.albumart_widget is not None:
                 self.albumart_widget.hide()
             self.unlock_dialog.hide()
             self.manager.unlock()
@@ -627,13 +619,13 @@ class Stage(Gtk.Window):
             self.unlock_dialog.blink()
 
     def authentication_cancel_callback(self, dialog):
-        self.cancel_unlock_widget()
+        self.manager.cancel_unlocking()
 
     def set_message(self, msg):
         """
         Passes along an away-message to the clock.
         """
-        if self.clock_widget != None:
+        if self.clock_widget is not None:
             self.clock_widget.set_message(msg)
 
     def initialize_pam(self):
@@ -652,9 +644,9 @@ class Stage(Gtk.Window):
 
         utils.clear_clipboards(self.unlock_dialog)
 
-        if self.clock_widget != None:
+        if self.clock_widget is not None:
             self.clock_widget.stop_positioning()
-        if self.albumart_widget != None:
+        if self.albumart_widget is not None:
             self.albumart_widget.stop_positioning()
 
         status.Awake = True
@@ -662,52 +654,46 @@ class Stage(Gtk.Window):
         if self.info_panel:
             self.info_panel.refresh_power_state()
 
-        if self.clock_widget != None:
+        if self.clock_widget is not None:
             self.clock_widget.show()
-        if self.albumart_widget != None:
+        if self.albumart_widget is not None:
             self.albumart_widget.show()
 
         self.unlock_dialog.show()
 
-        if self.audio_panel != None:
+        if self.audio_panel is not None:
             self.audio_panel.show_panel()
-        if self.info_panel != None:
-            self.info_panel.update_visibility()
-        if self.osk != None:
+        if self.info_panel is not None:
+            self.info_panel.refresh_power_state()
+        if self.osk is not None:
             self.osk.show()
 
     def cancel_unlocking(self):
         if self.unlock_dialog:
-            self.unlock_dialog.cancel_auth_client()
-
-    def cancel_unlock_widget(self):
-        """
-        Hide the unlock widget (and others) if the unlock has been canceled
-        """
-        if not status.Awake:
-            return
+            self.unlock_dialog.cancel()
 
         self.set_timeout_active(None, False)
         utils.clear_clipboards(self.unlock_dialog)
 
-        self.unlock_dialog.hide()
-
-        if self.clock_widget != None:
+        if self.unlock_dialog is not None:
+            self.unlock_dialog.hide()
+        if self.clock_widget is not None:
             self.clock_widget.hide()
-        if self.albumart_widget != None:
+        if self.albumart_widget is not None:
             self.albumart_widget.hide()
-        if self.audio_panel != None:
+        if self.audio_panel is not None:
             self.audio_panel.hide()
-        if self.info_panel != None:
+        if self.info_panel is not None:
             self.info_panel.hide()
-        if self.osk != None:
+        if self.osk is not None:
             self.osk.hide()
 
-        self.unlock_dialog.cancel()
         status.Awake = False
 
         self.update_monitor_views()
-        self.info_panel.update_visibility()
+
+        if self.info_panel is not None:
+            self.info_panel.refresh_power_state()
 
     def update_monitor_views(self):
         """
@@ -716,9 +702,9 @@ class Stage(Gtk.Window):
         """
 
         if not status.Awake:
-            if self.clock_widget != None and settings.get_show_clock():
+            if self.clock_widget is not None and settings.get_show_clock():
                 self.clock_widget.start_positioning()
-            if self.albumart_widget != None and settings.get_show_albumart():
+            if self.albumart_widget is not None and settings.get_show_albumart():
                 self.albumart_widget.start_positioning()
 
         for monitor in self.monitors:
@@ -762,8 +748,7 @@ class Stage(Gtk.Window):
         else:
             self.rect = status.screen.get_screen_geometry()
 
-        if status.Debug:
-            print("Stage.update_geometry - new backdrop position: %d, %d  new size: %d x %d" % (self.rect.x, self.rect.y, self.rect.width, self.rect.height))
+        DEBUG("Stage.update_geometry - new backdrop position: %d, %d  new size: %d x %d" % (self.rect.x, self.rect.y, self.rect.width, self.rect.height))
 
         hints = Gdk.Geometry()
         hints.min_width = self.rect.width
